#!/usr/bin/env python
""" cdecl.py - parse c declarations

(c) 2002, 2003 Simon Burton <simon@arrowtheory.com>
Released under GNU GPL license.

version 0.xx

"""
from __future__ import generators

import sys
import string
import types
import copy

# ==================================================== #

class Lexer:
  def __init__(self,s="",verbose=0,**kw):
    self.verbose = verbose
    # map keywords and typedefs
    self.lookup = {}
    for t in \
      "float double void char int".split():
      self.lookup[t] = BasicType( t )
    for t in "signed unsigned short long const volatile".split():
      self.lookup[t] = Qualifier( t )
    for t in "extern static register auto".split():
      self.lookup[t] = StorageClass( t )
    self.lookup['typedef'] = Typedef()
    self.stack = None
    self.lines = s.splitlines()
    self.set_state("","",0,0)
    self.so_file = ""
    self._newline()
    for key in kw.keys():
      self.__dict__[key] = kw[key]
    self.get_token() # start
  def mktypedef(self,tok,node):
    if self.verbose:
      print "%s.mktypedef(%s,%s)"%(self,tok,node)
    self.lookup[ tok ] = node
  def _get_kind(self,tok):
    #print '_get_kind(%s)'%tok,self.lookup
    try:
      return self.lookup[tok].clone()
    except KeyError:
      if tok.startswith("__builtin"):
        node = GCCBuiltin(tok)
        self.lookup[tok] = node
        return node
      return None
  def _newline(self):
    while self.lno < len(self.lines):
      line = self.lines[self.lno]
      if not line or line[0] != "#":
        break
      l = line.split('"')
      assert len(l)>=2
      self.so_file = l[1]
      #self.so_lno = int( l[0].split()[1] )
      #sys.stderr.write("# %s %s: %s\n"%(so_lno,so_file,l))
      self.lno+=1
  def get_token(self):
    self.push_state()
    ident_chars0 = string.letters+"_"
    ident_chars1 = string.letters+string.digits+"_"
    tok, kind = "", ""
    while self.lno < len(self.lines):
      s = self.lines[self.lno]
      i=self.col
      while i < len(s):
        if s[i].isspace():
          i=i+1
          continue
        #if s[i] in ident_chars0:
        if s[i].isalpha() or s[i]=='_':
          # identifier
          j=i+1
          while j<len(s):
            if s[j] in ident_chars1:
              j=j+1
            else:
              break
          tok = s[i:j]
          self.col = j
          kind = self._get_kind(tok)
          break
        if s[i].isdigit() or \
          (i+1<len(s) and s[i] in '+-.' and s[i+1].isdigit()):
          # number literal
          is_float = s[i]=='.'
          is_hex = s[i:i+2]=='0x'
          if is_hex:
            i=i+2
            assert s[i].isdigit()
          j=i+1
          while j<len(s):
            #print "lex ",repr(s[i]),is_float
            if s[j].isdigit():
              j=j+1
            elif s[j]=='.' and not is_float:
              assert not is_hex
              j=j+1
              is_float=1
            else:
              break 
          tok = s[i:j]
          self.col = j
          if is_float:
            kind = float(tok)
          elif is_hex:
            kind = int(tok,16)
          else:
            kind = int(tok)
          break
        if s[i:i+3]=='...':
          # ellipses
          #sys.stderr.write( "ELLIPSES "+str(self.get_state()) )
          tok = s[i:i+3]
          kind = tok
          self.col = i+3
          break
        if s[i] in '*{}()[]:;,=+-.<>|&':
          tok = s[i]
          kind = tok
          self.col = i+1
          break
        # keep moving
        #sys.stderr.write( "lexer ignoring '%s'\n"%s[i] )
        i=i+1
      if i==len(s):
        # nothing found
        assert tok == ""
        self.col=0
        self.lno+=1
        self._newline()
      else:
        assert tok
        break
    self.set_state(tok,kind,self.lno,self.col)
  def err_string(self):
    "Return helpful error string :)"
    return self.lines[self.lno]+"\n"+" "*self.col+"^\n"
  def push_state(self):
    self.stack = self.get_state() # short stack
    #self.stack.push( self.get_state() )
  def unget_token(self):
    assert self.stack is not None
    self.set_state(*self.stack)
    self.stack = None
  def set_state(self,tok,kind,lno,col):
    if self.verbose:
      print "tok,kind,lno,col = ",(tok,kind,lno,col)
    self.tok = tok
    self.kind = kind
    self.lno = lno # line
    self.col = col # column
  def get_state(self):
    return self.tok,self.kind,self.lno,self.col
  def get_file(self):
    return self.so_file

# ==================================================== #

class Symbols:
  def __init__(self,parent=None):
    self.verbose = verbose
    self.parent=parent # are we a nested namespace?
    self.lookup = {} # identifiers
    self.tags = {} # struct, union, enum tags
  def __str__(self):
    return "Symbols(%s,%s)"%(self.lookup,self.tags)
  #def new_child(self):
    #return Symbols(None)
    #return Symbols(self)
  def __getitem__(self,key):
    try:
      item = self.lookup[key]
    except KeyError:
      item = None
      #if self.parent is not None:
        #item = self.parent[item]
        ## self[key] = item # cache
    #if self.verbose: print "%s.get('%s')='%s'"%(self,key,item)
    return item
  def __setitem__(self,key,val):
    #if self.verbose: print "%s.set('%s','%s')"%(self,key,val)
    assert val is not None
    self.lookup[key] = val
  def set_tag(self,key,val):
    #if self.verbose: print "%s.set_tag(%s,%s)"%(self,key,val)
    assert len(key)
    self.tags[key] = val
  def deep_get_tag(self,key):
    try:
      item = self.tags[key]
    except KeyError:
      item = None
      if self.parent is not None:
        item = self.parent.deep_get_tag(key)
    #if self.verbose: print "%s.get_tag(%s)=%s"%(self,key,item)
    return item
  def get_tag(self,key):
    try:
      item = self.tags[key]
    except KeyError:
      item = None
    #if self.verbose: print "%s.get_tag(%s)=%s"%(self,key,item)
    return item

# ==================================================== #

class ParseError(Exception):
  def __init__(self,*e):
    self.e = e
  def __str__(self):
    return string.join(map(str,self.e))

class Parser:
  def parse_error(self,lexer,reason,*blah):
    sys.stderr.write( "%s.parse_error()\n"%self )
    sys.stderr.write( "at line %s: %s\n"%(lexer.lno+1,reason) )
    sys.stderr.write( lexer.err_string() )
    raise ParseError(reason,*blah)
  def expected_error(self,lexer,*l):
    self.parse_error( lexer, "expected %s, got '%s'"\
      %(string.join(map(repr,l)," or "),lexer.tok))
  def consume(self,lexer,tok):
    if lexer.tok != tok:
      self.expected_error(lexer, tok)
    lexer.get_token()
  def parse_enter(self,lexer):
    #return
    self.start_lno=lexer.lno
    self.file=lexer.so_file
  def parse_leave(self,lexer):
    #return
    self.lines =\
      lexer.lines[self.start_lno:max(lexer.lno,self.start_lno+1)]

# ==================================================== #

class Node(list):
  " A node in a parse tree "
  def __init__(self,*items,**kw):
    list.__init__( self, items )
    self.lock1 = 0
    self.lock2 = 0
    self.verbose = 0
    for key in kw.keys():
      self.__dict__[key] = kw[key]

  def __str__(self):
    return "%s%s"%(self.__class__.__name__,tuple(self))

  def __repr__(self):
    return str(self)

  def __eq__(self,other):
    if not isinstance(other,Node):
      return 0
    if len(self)!=len(other):
      return 0
    for i in range(len(self)):
      if not self[i]==other[i]:
        return 0
    return 1
  def __ne__(self,other):
    return not self==other
  def filter(self,klass):
    return [node for node in self if isinstance(node,klass)]
    #return filter( lambda x:isinstance(x,klass), self )

  def find(self,klass):
    for node in self:
      if isinstance(node,klass):
        return node
    return None

  def leaves(self):
    for i in self:
      if isinstance( i, Node ):
        for j in i.leaves():
          yield j
      else:
        yield i

  def nodes(self):
    for i in self:
      if isinstance( i, Node ):
        for j in i.nodes():
          yield j
    yield self

  def deeplen(self):
    i=0
    if not self.lock2:
      self.lock2=1
      for item in self:
        i+=1
        if isinstance(item,Node):
          i+=item.deeplen()
      self.lock2=0
    else:
      i+=1
    return i

  def deepstr(self,level=0):
    if self.deeplen() < 4:
      nl = ""; indent = ""
    else:
      nl="\n"; indent = "  "
    s = []
    if not self.lock1:
      self.lock1=1
      for item in self:
        if isinstance(item,Node):
          s.append( indent*(level+1)+item.deepstr(level+1) )
        else:
          s.append( indent*(level+1)+repr(item) )
      self.lock1=0
    else:
      for item in self:
        if isinstance(item,Node):
          s.append( indent*(level+1)+"..." )
        else:
          s.append( indent*(level+1)+"%s"%repr(item) )
    return "%s(%s)"%\
      (self.__class__.__name__,nl+string.join(s,","+nl))

  def clone(self):
    return copy.deepcopy(self)
    l = []
    for item in self:
      if isinstance(item,Node):
        item = item.clone()
      l.append(item)
    #raise blahthiswontworkerror
    return self.__class__(*l)

  def expose(self,klass):
    ' expose children of any <klass> instance '
    # children first
    for x in self:
      if isinstance(x,Node):
        x.expose(klass)
    # now the tricky bit
    i=0
    while i < len(self):
      if isinstance(self[i],klass):
        node=self.pop(i)
        for x in node:
          assert not isinstance(x,klass)
          if hasattr(node,'lines') and not hasattr(x,'lines'):
            x.lines=node.lines
          self.insert(i,x) # expose
          i=i+1
          assert i<=len(self)
      else:
        i=i+1

  def delete(self,klass):
    ' delete any <klass> subtree '
    for x in self:
      if isinstance(x,Node):
        x.delete(klass)
    # now the tricky bit
    i=0
    while i < len(self):
      if isinstance(self[i],klass):
        self.pop(i)
      else:
        i=i+1

  def explain(self):
    l = []
    for x in self:
      if isinstance(x,Node):
        l.append(x.explain())
      else:
        l.append(str(x))
    return string.join(l," ") 
      #(self.__class__.__name__,string.join(l) )

  def psource(self):
    if hasattr(self,'lines'):
      print "# "+string.join(self.lines,"\n# ")+"\n"

  def report(self):
    self.psource()
    print self.deepstr()

  def ctype(self):
    l = []
    for x in self:
      if isinstance(x,Node):
        l.append(x.ctype())
      else:
        l.append(x)
    return self.__class__(*l)

  def cstr(self,l=None):
    #print '%s.cstr(%s)'%(self,l)
    if l is None:
      l = []
    for x in self:
      if isinstance(x,Node):
        x.cstr(l)
      else:
        l.insert(0,str(x)+' ')
    return string.join(l,'')

# ==================================================== #

class BasicType(Node):
  " int char short etc. "
  def __init__(self,name):
    Node.__init__(self,name)

class Qualifier(Node):
  def __init__(self,name):
    Node.__init__(self,name)
    self.name=name

class StorageClass(Node):
  def __init__(self,name):
    Node.__init__(self,name)
    self.name=name

class Typedef(StorageClass):
  def __init__(self):
    Node.__init__(self,'typedef')
  def explain(self):
    return "type"
  #def cstr(self,l):
    #l.insert

class GCCBuiltin(BasicType):
  pass

class Identifier(Node):
  def __init__(self,*items):
    Node.__init__(self,*items)
    self.name=items[0]
  #def __str__(self):
    #return "Ident('%s')"%self.name
  def explain(self):
    if len(self)==1:
      return "%s"%self.name
    else:
      return "%s initialized to %s"%(self.name,
        Node(self[1]).explain()) # will handle Initializer
  def ctype(self):
    #assert len(self)==2
    return self.__class__(*self[1:]) #.clone() ?
  def cstr(self,l=None):
    if l is None:
      l=[]
    if len(self)>1:
      assert len(self)==2
      l.append( '%s = %s'%(self[0],self[1]) )
    else:
      l.append( str(self[0]) )
    return string.join(l)

class Function(Node,Parser):
  def __init__(self,*items):
    Node.__init__(self,*items)
  def parse(self,lexer,symbols):
    symbols = Symbols(symbols)
    args = ''
    #lexer.get_token()
    if lexer.tok != ')':
      if not lexer.tok: self.parse_error(lexer)
      #lexer.unget_token() # unget start of decl
      while lexer.tok != ')':
        node = ParameterDeclaration()
        node.parse(lexer,symbols)
        self.append( node )
        if lexer.tok != ')' and lexer.tok != ',':
          self.parse_error(lexer)
        if lexer.tok == ',':
          lexer.get_token()
    lexer.get_token()
  def explain(self):
    if len(self):
      return "function (%s), returning"%\
        string.join( map(lambda x:x.explain(),self), ", " )
    else:
      return "function returning"
  def cstr(self,l):
    #print '%s.cstr(%s)'%(self,l)
    _l=[]
    assert len(self)
    i=0
    while isinstance(self[i],Declarator):
      _l.append( self[i].cstr() )
      i=i+1
    l.append( '(%s)'%string.join(_l, ',') )
    while i<len(self):
      self[i].cstr(l)
      i=i+1
    return string.join(l)

class Pointer(Node):
  def __init__(self):
    Node.__init__(self)
  def explain(self):
    return "pointer to"
  def cstr(self,l):
    assert len(self)
    node=self[0]
    l.insert(0,'*')
    if isinstance(node,Function):
      l.insert(0,'(')
      l.append(')')
    elif isinstance(node,Array):
      l.insert(0,'(')
      l.append(')')
    return Node.cstr(self,l)

class Array(Node,Parser):
  def __init__(self,*items):
    Node.__init__(self,*items)
  def parse(self,lexer,symbols):
    lexer.get_token() # a number or ']'
    # XXX hack hack
    while lexer.tok != ']':
      self.append( lexer.kind )
      lexer.get_token()
    lexer.get_token() # read past the ']'
  def explain(self):
    s=''
    if len(self):
      if type(self[0])==int:
        s='0 to %s '%(self[0]-1)
    return "array %sof"%s
  def cstr(self,l):
    if len(self)==1:
      l.append('[]')
    elif len(self)==2:
      l.append('[%s]'%self[1])
    else:
      assert 0
    return self[0].cstr(l)

class Tag(Node):
  pass

class Compound(Node,Parser):
  "Struct or Union"
  def __init__(self,*items,**kw):
    Node.__init__(self,*items,**kw)
  def parse(self,lexer,symbols):
    symbols = Symbols(symbols)
    #lexer.get_token()
    tag = "" # anonymous
    if lexer.tok != '{':
      tag = lexer.tok
      if not ( tag[0]=='_' or tag[0].isalpha() ):
        self.parse_error(lexer ,"expected tag, got '%s'"%tag )
      lexer.get_token()
    if tag:
      self.append(Tag(tag))
    else:
      self.append(Tag())
    self.tag = tag
    if lexer.tok == '{':
      fieldlist = []
      lexer.get_token()
      if lexer.tok != '}':
        if not lexer.tok: self.parse_error(lexer)
        while lexer.tok != '}':
          node = StructDeclaration()
          node.parse(lexer,symbols)
          fieldlist.append( node )
      self += fieldlist
      lexer.get_token()
    if self.verbose: 
      print "%s.__init__() #<--"%(self)
  def cstr(self,_l=None):
    assert isinstance( self[0], Tag )
    tag=''
    if len(self[0]):
      tag=' '+self[0][0]
    if isinstance(self,Struct):
      l=[ 'struct%s '%tag ]
    elif isinstance(self,Union):
      l=[ 'union%s '%tag ]
    if len(self)>1:
      l.append('\n{\n')
      for node in self[1:]:
        l.append( node.cstr()+';\n' )
      l.append('} ')
    if _l is None:
      _l=[]
    while l:
      _l.insert( 0, l.pop() )
    return string.join( _l, '' )

class Struct(Compound):
  def explain(self):
    tag=self.tag
    if len(tag): tag=' '+tag
    if len(self)>1:
      return "struct%s containing %s"%\
        ( tag,
        string.join( map(lambda x:x.explain(),self[1:])," and " ) )
    else:
       return "struct%s"%tag

class Union(Compound):
  def explain(self):
    if len(self)>1:
      return "union %s containing %s"%\
        ( self.tag,
        string.join( map(lambda x:x.explain(),self[1:])," and " ) )
    else:
       return "union %s"%self.tag

class Enum(Node,Parser):
  def __init__(self,*items,**kw):
    Node.__init__(self,*items,**kw)
  def parse(self,lexer,symbols):
    tag = "" # anonymous
    if lexer.tok != '{':
      tag = lexer.tok
      if not ( tag[0]=='_' or tag[0].isalpha() ):
        self.parse_error(lexer ,"expected tag, got '%s'"%tag )
      lexer.get_token()
    if tag:
      self.append(Tag(tag))
    else:
      self.append(Tag())
    self.tag = tag
    if lexer.tok == '{':
      lexer.get_token()
      if lexer.tok != '}':
        if not lexer.tok:
          self.parse_error(lexer)
        while lexer.tok != '}':
          if lexer.kind is not None:
            self.expected_error(lexer ,"identifier" )
          ident = Identifier(lexer.tok)
          if symbols[ident[0]] is not None:
            self.parse_error(lexer,"%s already defined."%ident[0])
          symbols[ident[0]]=ident
          self.append( ident )
          lexer.get_token()
          if lexer.tok == '=':
            lexer.get_token()
            # ConstantExpr
            if type( lexer.kind ) is not int:
              #self.parse_error(lexer ,"expected integer" )
              # XXX hack hack
              while lexer.tok!=',' and lexer.tok!='}':
                lexer.get_token()
            else:
              # put initializer into the Identifier
              ident.append( lexer.kind )
              lexer.get_token()
          if lexer.tok != '}':
            if lexer.tok != ',':
              self.expected_error(lexer,"}",",")
            lexer.get_token() # ','
      lexer.get_token()
    if self.verbose:
      print "%s.__init__() #<--"%(self)
  def explain(self):
    tag = self.find(Tag)
    tag_s=""
    if len(tag):
      tag_s = " "+tag[0]
    if len(self)>1:
      #l=[]
      #for x in self[1:]: # skip Tag
       # 
      return "enumeration%s of %s"%\
        ( tag_s,
        string.join( map(lambda x:x.explain(),self[1:]),", then " ) )
    else:
       return "enumeration%s"%tag_s
  def cstr(self,_l=None):
    assert isinstance( self[0], Tag )
    tag=''
    if len(self[0]):
      tag=' '+self[0][0]
    l=[ 'enum%s'%tag ]
    if len(self)>1:
      l.append('\n{\n')
      for node in self[1:]:
        l.append( node.cstr()+',\n' )
      l.append('} ')
    if _l is None:
      _l=[]
    while l:
      _l.insert( 0, l.pop() )
    return string.join( _l, '' )

class Declarator(Node,Parser):
  def __init__(self,*items):
    Node.__init__(self,*items)
    self.ident = None
  def parse(self,lexer,symbols):
    #Parser.parse_enter(self,lexer)
    stack = []
    # read up to identifier, pushing tokens onto stack
    self.ident = self._identifier(lexer,symbols,stack)
    self.name = ''
    if self.ident is not None:
      self.append( self.ident )
      self.name = self.ident.name
    # now read outwards from identifier
    self._declarator(lexer,symbols,stack)
    #Parser.parse_leave(self,lexer)
  def _identifier(self,lexer,symbols,stack):
    if self.verbose:
      print "%s._identifier()"%self
    ident = None
    if lexer.tok != ';':
      while lexer.tok and lexer.kind is not None:
        stack.append( (lexer.tok, lexer.kind) )
        lexer.get_token()
      if lexer.tok:
        ident = Identifier( lexer.tok )
        #stack.append( (ident.name, ident) )
        lexer.get_token()
    if self.verbose:
      print "%s._identifier()=%s"%(self,repr(ident))
    return ident
  def _declarator(self,lexer,symbols,stack,level=0):
    if self.verbose:
      print "  "*level+"%s._declarator(%s) # --->"%\
        (self,stack)
    if lexer.tok == '[':
      while lexer.tok == '[':
        node = Array()
        node.parse(lexer,symbols)
        self.append(node)
      if lexer.tok == '(':
        self.parse_error(lexer ,"array of functions" )
    elif lexer.tok == '(':
      lexer.get_token()
      node = Function()
      node.parse(lexer,symbols)
      self.append( node )
      if lexer.tok == '(':
        self.parse_error(lexer ,"function returns a function" )
      if lexer.tok == '[':
        self.parse_error(lexer ,"function returns an array" )
    while stack:
      tok, kind = stack[-1] # peek
      if tok == '(':
        stack.pop()
        self.consume(lexer,')')
        #if lexer.tok != ')':
          #self.expected_error(lexer,')')
        #lexer.get_token()
        self._declarator(lexer,symbols,stack,level+1)
      elif tok == '*':
        stack.pop()
        self.append( Pointer() )
      else:
        tok, kind = stack.pop()
        self.append( kind )
    if self.verbose:
      print "  "*level+"%s._declarator(%s) # <---"%\
        (self,stack)
  def specify(self,spec):
    'subsume our specifiers'
    for i in range(len(spec)):
      self.append(spec[i])
    #self.append(spec)
  def is_function(self):
    return len(self)>1 and isinstance(self[1],Function)
  def is_typedef(self):
    #return self.filter(Typedef)
    declspec = self.filter(DeclarationSpecifiers)[0]
    return declspec.filter(Typedef)
    ##return filter( lambda x: x=='typedef', self.leaves() )
  def act(self):
    # children go first 
    for node in self.nodes():
      if isinstance(node,Type):
        node.do_type()
    self.expose(Type)
  def explain(self):
    if len(self):
      if isinstance( self[0], Identifier ):
        return "%s is %s"%(self[0].explain(),
          string.join(map(lambda x:x.explain(),self[1:])," "))
      else:
        return Node.explain(self)
    return ''
  #def cstr(self,l=None):
    #return Node.cstr(self,l)+';'

class AbstractDeclarator(Declarator):
  """ used in ParameterDeclaration; may lack an identifier """
  def _identifier(self,lexer,symbols,stack):
    if self.verbose:
      print "%s._identifier()"%self
    ident = None
    while 1:
      if lexer.tok == ';':
        self.parse_error(lexer)
      if lexer.tok == ')':
        break
      if lexer.tok == ',':
        break
      if lexer.kind is None:
        #print "%s.new identifier"%self
        ident = Identifier( lexer.tok )
        lexer.get_token()
        #stack.append( (ident.name, ident) )
        break
      stack.append( (lexer.tok, lexer.kind) )
      lexer.get_token()
    if self.verbose:
      print "%s._identifier()=%s"%(self,repr(ident))
    return ident
  #def cstr(self,l=None):
    #return Node.cstr(self,l)

class FieldLength(Node):
  def explain(self):
    return ""
  def cstr(self,l):
    l.append(':%s'%self[0]) 

class StructDeclarator(Declarator):
  def parse(self,lexer,symbols):
    if lexer.tok != ':':
      Declarator.parse(self,lexer,symbols)
    if lexer.tok == ':':
      lexer.get_token()
      # ConstantExpr
      length = int(lexer.tok)
      #print "length = ",length
      self.append( FieldLength(length) )
      lexer.get_token()
  def explain(self):
    flen = self.find(FieldLength)
    if flen is not None:
      i = self.index(flen)
      self.pop(i)
      s = Declarator.explain(self)
      self.insert(i,flen)
      width = flen[0]
      if width > 0:
        return s+" bitfield %s wide"%width
      else:
        return s+" alignment bitfield"
    else:
      return Declarator.explain(self)

class DeclarationSpecifiers(Node,Parser):
  def __init__(self,*items):
    Node.__init__(self,*items)
  def parse(self,lexer,symbols):
    self._spec(lexer,symbols)
    self.reverse()
  def _spec(self,lexer,symbols):
    typespec = None
    while lexer.tok:
      if isinstance( lexer.kind, Type ) or\
        isinstance( lexer.kind, BasicType ):
        if typespec is not None:
          self.parse_error(lexer ,"type already specified as %s"\
            %typespec )
        typespec=lexer.kind
        self.append( lexer.kind )
        lexer.get_token()
      elif isinstance( lexer.kind, Qualifier ):
        self.append( lexer.kind )
        lexer.get_token()
      elif isinstance( lexer.kind, StorageClass ):
        self.append( lexer.kind )
        lexer.get_token()
      elif lexer.tok=='struct':
        lexer.get_token()
        self._struct(lexer,symbols)
        break #?
      elif lexer.tok=='union':
        lexer.get_token()
        self._union(lexer,symbols)
        break #?
      elif lexer.tok=='enum':
        lexer.get_token()
        self._enum(lexer,symbols)
        break #?
      elif lexer.kind is None:
        # identifier
        break
      else:
        break
  def _struct(self,lexer,symbols):
    if self.verbose:
      print "%s._struct()"%(self)
    node = Struct()
    node.parse(lexer,symbols)
    _node = None
    if node.tag:
      _node = symbols.get_tag( node.tag )
    if _node is not None:
      if not isinstance( _node, Struct ):
        self.parse_error(lexer,"tag defined as wrong kind")
      if len(node)>1:
        if len(_node)>1:
          self.parse_error(lexer,"tag already defined as %s"%_node)
        #symbols.set_tag( node.tag, node )
      #else:
        # refer to the previously defined struct
        ##node = _node
        #node = _node.clone()
    if 0:
      # refer to the previously defined struct
      if len(node)==1:
        _node = symbols.deep_get_tag( node.tag )
        if _node is not None:
          node=_node
      # what about any future definition of this struct ?
    if node.tag:
      symbols.set_tag( node.tag, node )
    self.append( node )
  def _union(self,lexer,symbols):
    if self.verbose:
      print "%s._union(%s)"%(self,node)
    node = Union()
    node.parse(lexer,symbols)
    _node = None
    if node.tag:
      _node = symbols.get_tag( node.tag )
    if _node is not None:
      if not isinstance( _node, Union ):
        self.parse_error(lexer,"tag %s defined as wrong kind"%repr(node.tag))
      if len(node)>1:
        if len(_node)>1:
          self.parse_error(lexer,"tag already defined as %s"%_node)
        #symbols.set_tag( node.tag, node )
      #else:
        #node = _node
    #if len(node)==1:
      #_node = symbols.deep_get_tag( node.tag )
      #if _node is not None:
        #node=_node
    if node.tag:
      symbols.set_tag( node.tag, node )
    self.append( node )
  def _enum(self,lexer,symbols):
    if self.verbose:
      print "%s._enum(%s)"%(self,node)
    node = Enum()
    node.parse(lexer,symbols)
    _node = None
    if node.tag:
      _node = symbols.get_tag( node.tag )
    if _node is not None:
      if not isinstance( _node, Enum ):
        self.parse_error(lexer,"tag defined as wrong kind")
      if len(node)>1:
        if len(_node)>1:
          self.parse_error(lexer,"tag already defined as %s"%_node)
        #symbols.set_tag( node.tag, node )
      #else:
        #node = _node
    #if len(node)==1:
      #_node = symbols.deep_get_tag( node.tag )
      #if _node is not None:
        #node=_node
    if node.tag:
      symbols.set_tag( node.tag, node )
    self.append( node )
  def is_typedef(self):
    return self.find(Typedef) is not None
  def needs_declarator(self):
    for node in self:
      if isinstance( node, Struct ):
        return False 
      if isinstance( node, Enum ):
        return False 
      if isinstance( node, Union ):
        return False 
    return True

class TypeSpecifiers(DeclarationSpecifiers):
  " used in ParameterDeclaration "
  def _spec(self,lexer,symbols):
    typespec = None
    while lexer.tok:
      if isinstance( lexer.kind, Type ) or\
        isinstance( lexer.kind, BasicType ):
        if typespec is not None:
          self.parse_error(lexer ,"type already specified as %s"\
            %typespec )
        typespec=lexer.kind
        self.append( lexer.kind )
        lexer.get_token()
      elif isinstance( lexer.kind, Qualifier ):
        self.append( lexer.kind )
        lexer.get_token()
      elif isinstance( lexer.kind, StorageClass ):
        self.parse_error(lexer ,"'%s' cannot appear here"%lexer.tok )
      elif lexer.tok=='struct':
        lexer.get_token()
        self._struct(lexer,symbols)
        break #?
      elif lexer.tok=='union':
        lexer.get_token()
        self._union(lexer,symbols)
        break #?
      elif lexer.tok=='enum':
        lexer.get_token()
        self._enum(lexer,symbols)
        break #?
      elif lexer.kind is None:
        # identifier
        break
      else:
        break

class Initializer(Node,Parser):
  def __init__(self,*items):
    Node.__init__(self,*items)
  def parse(self,lexer,symbols):
    self.parse_error(lexer,"not implemented")

class Type(Node):
  " typedefed things "
  def __init__(self,name,dcln):
    Node.__init__(self,name,dcln)
    self.name=name
    #self.dcln=dcln
  def do_type(self):
    assert type(self.pop(0))==str
    self.expose(Declarator)
    self.delete(Identifier)
    self.delete(Typedef)

class Declaration(Node,Parser):
  def __init__(self,*items):
    Node.__init__(self,*items)
    #self.acted=False
  def parse(self,lexer,symbols):
    if not lexer.tok: 
      return
    Parser.parse_enter(self,lexer)
    declspec = DeclarationSpecifiers() 
    declspec.parse(lexer,symbols)
    if len(declspec)==0:
      if lexer.tok == ';':
        lexer.get_token()
        # empty declaration...
        return
      self.parse_error(lexer,
        "expected specifiers, got '%s'"%lexer.tok )
    self.append(declspec)
    while 1:
      decl = Declarator()
      decl.parse(lexer,symbols)
      if len(decl)==0:
        if declspec.needs_declarator():
          self.parse_error(lexer,
            "expected declarator, got '%s'"%lexer.tok )
      #decl.specify(declspec)
      self.append(decl)
      ident = decl.ident
      if ident is not None:
        # install symbol
        node = symbols[ident[0]]
        if node is not None:
          self.parse_error(lexer,
            "%s already defined as %s"%(ident,node))
        else:
          if self.is_typedef():
            #lexer.mktypedef( ident[0], self )
            tp = Type(ident[0],self) 
            lexer.mktypedef( ident[0], tp )
          else:
            symbols[ident[0]] = self
        if lexer.tok == '=':
          # parse initializer
          lexer.get_token()
          init = Initializer()
          init.parse(lexer,symbols)
          ident.append( init ) # as in Enum
      #else: struct, union or enum
      if lexer.tok == ';':
        # no more declarators
        break
      self.consume(lexer,',')
    self.consume(lexer,';')
    Parser.parse_leave(self,lexer)
  def is_typedef(self):
    return self[0].is_typedef()
  def do_spec(self):
    " distribute DeclarationSpecifiers over each Declarator "
    spec=self[0]
    assert isinstance(spec,DeclarationSpecifiers)
    self.pop(0)
    for decl in self:
      assert isinstance(decl,Declarator)
      for node in spec:
        decl.append(node.clone())
  def act(self):
    #assert not self.acted
    # children go first 
    for node in self.nodes():
      if isinstance(node,Declaration):
        node.do_spec()
    #self.acted=True
    self.expose(Declaration)
  def explain(self):
    return string.join([x.explain() for x in self],", ") 
    return string.join(map(lambda x:x.explain(),self),", ") 

class ParameterDeclaration(Declaration):
  def parse(self,lexer,symbols):
    typespec = TypeSpecifiers()
    typespec.parse(lexer,symbols)
    self.append(typespec)
    decl = AbstractDeclarator()
    decl.parse(lexer,symbols)
    self.append(decl)
    #decl.specify(typespec)
    ident = decl.ident
    if ident is not None:
      node = symbols[ident[0]]
      if node is not None:
        self.parse_error(lexer,
          "%s already defined as %s"%(ident,node))
      else:
        symbols[ident[0]] = self

class StructDeclaration(Declaration):
  def parse(self,lexer,symbols):
    if not lexer.tok: 
      return
    declspec = DeclarationSpecifiers() 
    declspec.parse(lexer,symbols)
    self.append(declspec)
    if len(declspec)==0:
      if lexer.tok == ';':
        lexer.get_token()
        # empty declaration...
        return
      self.parse_error(lexer,
        "expected specifiers, got '%s'"%lexer.tok )
    while 1:
      decl = StructDeclarator()
      decl.parse(lexer,symbols)
      if len(decl)==0:
        self.parse_error(lexer,
          "expected declarator, got '%s'"%lexer.tok )
      self.append(decl)
      #decl.specify(declspec)
      ident = decl.ident
      if ident is not None:
        node = symbols[ident[0]]
        if node is not None:
          self.parse_error(lexer ,
            "%s already defined as %s"%(ident,node))
        else:
          if declspec.is_typedef():
            self.parse_error(lexer,"typedef in struct or union")
          else:
            symbols[ident[0]] = self
      if lexer.tok == ';':
        break
      self.consume(lexer,',')
    self.consume(lexer,';')

class Rewrite:
  " visit nodes in tree... "
  def __init__(self,node=None):
    if node is not None:
      self.visit(node)
  def visit(self,node):
    #print 'visit(%s)'%node
    for _node in node:
      if isinstance(_node,Declarator):
        self.visit_decl(_node)
      elif isinstance(_node,Node):
        self.visit(_node)
  def visit_decl(self,decl):
    assert isinstance(decl,Declarator)
    i=len(decl)
    accum=[] # accumulate nodes
    while i:
      i=i-1
      node=decl.pop(i)
      if isinstance(node,Declarator):
        self.visit_decl(node)
      else:
        self.visit(node)
      if isinstance(node,Pointer):
        node+=accum
        accum=[node]
      elif isinstance(node,Function):
        node+=accum
        accum=[node]
      elif isinstance(node,Array):
        while accum:
          node.insert(0,accum.pop())
        accum=[node]
        # array size (if any) at end
      #elif isinstance(node,Identifier):
        #node+=accum
        #accum=[node]
      else:
        # accumulate
        accum.insert(0,node)
    decl[:]=accum

class File(Node,Parser):
  def __init__(self,*items,**kw):
    Node.__init__(self,*items,**kw)
  def parse(self,s):
    self.symbols = Symbols()
    lexer = Lexer(s,verbose=verbose)
    node = None
    while lexer.tok:
      node=Declaration()
      node.parse(lexer,self.symbols)
      if node:
        self.append(node)
        #node.psource()
        #print node.deepstr(),'\n'
        #node.act()
  def strip(self,files):
    ' leave only declarations in <files> '
    i=0
    while i<len(self):
      if self[i].file in files:
        i=i+1
      else:
        self.pop(i)
  def assert_no_dups(self):
    check={}
    for node in self.nodes():
      assert not check.has_key(id(node))
      check[id(node)]=1
  def act(self, verbose=0):
    rw=Rewrite()
    i=0
    while i < len(self):
      if verbose: print "##"*25
      node=self[i]
      if verbose: node.psource()
      if verbose: print node.deepstr(),'\n'
      assert isinstance(node,Declaration)
      #try:
      node.act()
      #except Exception, e:
        #node.report()
        #raise e
      #print node.deepstr(),'\n'
      self[i:i+1] = node # expose declarators from node
      for j in range(len(node)):
        node=self[i]
        assert isinstance(node,Declarator)
        node.act()
        if verbose: print node.deepstr(),'\n'
        rw.visit_decl(node)
        if verbose: print node.deepstr(),'\n'
        #print node.ctype().deepstr(),'\n'
        #assert node.clone() == node
        if 0:
          cstr = node.cstr()+';'
          if verbose: print '# '+cstr.replace('\n','\n# ')
          #print
          declaration = Declaration()
          declaration.parse(  Lexer( cstr ), Symbols() )
          declaration.act()
          assert len(declaration)==1
          decl=declaration[0]
          decl.act()
          rw.visit_decl(decl)
          if decl!=node:
            if verbose: print "#???????????"
            if verbose: print decl[0].deepstr(),'\n\n'
            assert 0
          if verbose: print '#OK'
        i=i+1

# ==================================================== #

try:
  import NoModule
  import psyco
  from psyco.classes import *
except ImportError:
  class _psyco:
    def jit(self):      pass
    def bind(self, f):  pass
    def proxy(self, f): return f
  psyco = _psyco()
psyco.bind( Lexer.get_token )
psyco.bind( Node )

verbose = 0
if __name__=="__main__":
  verbose = 0
  if not sys.argv[1:]:
    s = sys.stdin.read()
  #elif sys.argv[1]=='explain':
    #explain = 1
    #sys.argv[1:] = sys.argv[2:]
  if sys.argv[1:]:
    #s = string.join(sys.argv[1:])
    s = sys.argv[1]
    if sys.argv[2:]:
      verbose = int(sys.argv[2])
  if 0:
    import profile
    profile.run('File(s)','prof.out')
    import pstats
    p=pstats.Stats('prof.out')
    p.strip_dirs().sort_stats(-1).print_stats()
  else:
    node = File(verbose = 1 )
    node.parse(s)
    node.act(1)
